@@ -56,8 +56,6 @@ static void at_cbm_notify(GAtResult *result, gpointer user_data)
 	unsigned char pdu[88];
 	long hexpdulen;
 
-	dump_response("at_cbm_notify", TRUE, result);
-
 	g_at_result_iter_init(&iter, result);
 
 	if (!g_at_result_iter_next(&iter, "+CBM:"))
@@ -66,6 +64,11 @@ static void at_cbm_notify(GAtResult *result, gpointer user_data)
 	if (!g_at_result_iter_next_number(&iter, &pdulen))
 		return;
 
+	if (pdulen > 88) {
+		ofono_error("Got a CBM message bigger than maximum size!");
+		return;
+	}
+
 	hexpdu = g_at_result_pdu(result);
 
 	if (!hexpdu) {
@@ -73,7 +76,7 @@ static void at_cbm_notify(GAtResult *result, gpointer user_data)
 		return;
 	}
 
-	ofono_debug("Got new Cell Broadcast via CBM: %s, %d", hexpdu, pdulen);
+	DBG("Got new Cell Broadcast via CBM: %s, %d", hexpdu, pdulen);
 
 	if (decode_hex_own_buf(hexpdu, -1, &hexpdulen, 0, pdu) == NULL) {
 		ofono_error("Unable to hex-decode the PDU");
@@ -94,7 +97,6 @@ static void at_cscb_set_cb(gboolean ok, GAtResult *result, gpointer user_data)
 	ofono_cbs_set_cb_t cb = cbd->cb;
 	struct ofono_error error;
 
-	dump_response("cscb_set_cb", ok, result);
 	decode_at_error(&error, g_at_result_final_response(result));
 
 	cb(&error, cbd->data);
@@ -139,9 +141,9 @@ static void at_cbs_clear_topics(struct ofono_cbs *cbs,
 		goto error;
 
 	if (data->cscb_mode_1)
-		sprintf(buf, "AT+CSCB=1,\"0-65535\"");
+		snprintf(buf, sizeof(buf), "AT+CSCB=1,\"0-65535\"");
 	else
-		sprintf(buf, "AT+CSCB=0,\"\"");
+		snprintf(buf, sizeof(buf), "AT+CSCB=0,\"\"");
 
 	if (g_at_chat_send(data->chat, buf, none_prefix,
 				at_cscb_set_cb, cbd, g_free) > 0)
@@ -180,8 +182,6 @@ static void at_cscb_support_cb(gboolean ok, GAtResult *result, gpointer user)
 	GAtResultIter iter;
 	char buf[256];
 
-	dump_response("at_cscb_support_cb", ok, result);
-
 	if (!ok)
 		goto error;
 
@@ -204,9 +204,9 @@ static void at_cscb_support_cb(gboolean ok, GAtResult *result, gpointer user)
 	 * of new topics using CSCB mode 0.
 	 */
 	if (data->cscb_mode_1)
-		sprintf(buf, "AT+CSCB=1,\"0-65535\"");
+		snprintf(buf, sizeof(buf), "AT+CSCB=1,\"0-65535\"");
 	else
-		sprintf(buf, "AT+CSCB=0,\"\"");
+		snprintf(buf, sizeof(buf), "AT+CSCB=0,\"\"");
 
 	if (g_at_chat_send(data->chat, buf, none_prefix,
 				at_cbs_register, cbs, NULL) > 0)
@@ -236,6 +236,11 @@ static int at_cbs_probe(struct ofono_cbs *cbs, unsigned int vendor,
 
 static void at_cbs_remove(struct ofono_cbs *cbs)
 {
+	struct cbs_data *data = ofono_cbs_get_data(cbs);
+
+	ofono_cbs_set_data(cbs, NULL);
+
+	g_free(data);
 }
 
 static struct ofono_cbs_driver driver = {
