@@ -186,8 +186,10 @@ void ofono_cbs_notify(struct ofono_cbs *cbs, const unsigned char *pdu,
 	if (cbs->assembly == NULL)
 		return;
 
-	if (!cbs->powered)
+	if (!cbs->powered) {
+		ofono_error("Ignoring CBS because powered is off");
 		return;
+	}
 
 	if (!cbs_decode(pdu, pdu_len, &c)) {
 		ofono_error("Unable to decode CBS PDU");
@@ -329,7 +331,7 @@ static void cbs_set_topics_cb(const struct ofono_error *error, void *data)
 		g_slist_free(cbs->new_topics);
 		cbs->new_topics = NULL;
 
-		ofono_debug("Setting Cell Broadcast topics failed");
+		DBG("Setting Cell Broadcast topics failed");
 		__ofono_dbus_pending_reply(&cbs->pending,
 					__ofono_error_failed(cbs->pending));
 		return;
@@ -397,7 +399,8 @@ static void cbs_set_powered_cb(const struct ofono_error *error, void *data)
 	DBusMessage *reply;
 
 	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
-		ofono_debug("Setting Cell Broadcast topics failed");
+		ofono_error("Setting Cell Broadcast topics failed");
+
 		if (!cbs->pending)
 			return;
 
@@ -437,7 +440,7 @@ static DBusMessage *cbs_set_powered(struct ofono_cbs *cbs, gboolean value,
 	if (cbs->powered == value)
 		goto reply;
 
-	if (!cbs->driver->set_topics)
+	if (!cbs->driver->set_topics || !cbs->driver->clear_topics)
 		goto done;
 
 	if (msg)
@@ -717,6 +720,13 @@ static void cbs_got_file_contents(struct ofono_cbs *cbs)
 
 	powered = g_key_file_get_boolean(cbs->settings, SETTINGS_GROUP,
 						"Powered", &error);
+
+	if (error) {
+		powered = TRUE;
+		g_key_file_set_boolean(cbs->settings, SETTINGS_GROUP,
+					"Powered", powered);
+	}
+
 	cbs_set_powered(cbs, powered, NULL);
 }
 
@@ -760,7 +770,7 @@ static void sim_cbmi_read_cb(int ok, int length, int record,
 		return;
 
 	str = cbs_topic_ranges_to_string(cbs->efcbmi_contents);
-	ofono_debug("Got cbmi: %s", str);
+	DBG("Got cbmi: %s", str);
 	g_free(str);
 }
 
@@ -807,7 +817,7 @@ static void sim_cbmir_read_cb(int ok, int length, int record,
 		return;
 
 	str = cbs_topic_ranges_to_string(cbs->efcbmir_contents);
-	ofono_debug("Got cbmir: %s", str);
+	DBG("Got cbmir: %s", str);
 	g_free(str);
 }
 
@@ -822,10 +832,10 @@ static void sim_cbmid_read_cb(int ok, int length, int record,
 	GSList *contents = NULL;
 
 	if (!ok)
-		return;
+		goto done;
 
 	if ((length % 2) == 1 || length < 2)
-		return;
+		goto done;
 
 	cbs->efcbmid_length = length;
 
@@ -845,14 +855,15 @@ static void sim_cbmid_read_cb(int ok, int length, int record,
 	}
 
 	if (contents == NULL)
-		return;
+		goto done;
 
 	cbs->efcbmid_contents = g_slist_reverse(contents);
 
 	str = cbs_topic_ranges_to_string(cbs->efcbmid_contents);
-	ofono_debug("Got cbmid: %s", str);
+	DBG("Got cbmid: %s", str);
 	g_free(str);
 
+done:
 	cbs_got_file_contents(cbs);
 }
 
@@ -861,7 +872,7 @@ static void cbs_got_imsi(struct ofono_cbs *cbs)
 	const char *imsi = ofono_sim_get_imsi(cbs->sim);
 	char *topics_str;
 
-	ofono_debug("Got IMSI: %s", imsi);
+	DBG("Got IMSI: %s", imsi);
 
 	cbs->settings = storage_open(imsi, SETTINGS_STORE);
 	if (cbs->settings == NULL)
